/* crc check program */
/* usage crc file1 file2 ... */
/* file name can be ambiguous */
/* for Aztec C86 v. 3.20e */
/* by Jon Dart, Mar. 1986 */

#define Version "1.2 (10-Apr-86)"

#include "stdio.h"
#include "fixpath.h"
#include "sgtty.h"
#include "ctype.h"
#include "truth.h"
#include "ascii.h"

#define PATHSIZE  65

extern char *malloc();
extern char *rindex();
extern FILE *fopen();

docrc(fn)
char *fn;
/* compute crc for 1 file, given the file name */
{
#define swap(x) ((x*256)+(x/256))
#define lo(x) (x & 0xFF)

    register unsigned int aa;
    FILE *fd;
    unsigned int c;
    char *fnp;

    if ((fd=fopen(fn,"r"))!=NULL) {
    /* make fnp point to last part of file name, after any directory
       or drive prefix: */
        fnp = rindex(fn,'\\');
        if (fnp==NULL) {
            if ((fnp = rindex(fn,':'))==NULL)
                fnp = fn;
            else
                fnp++;
        }
        else
            fnp++;
        printf("%s ",fnp);
        /* compute the crc. (Ref: Dr. Dobb's Journal, Feb. 1986, p. 80) */
        aa = 0;
        while ((c=getc(fd)) != -1) {
             aa = swap(aa) ^ c;
             aa = aa ^ (lo(aa) >> 4);
             aa = aa ^ (swap(lo(aa)) << 4) ^ (lo(aa) << 5);
        }
        if (ferror(fd)) {
             printf("Read error on file\n");
             if (!isatty(stdout)) fprintf(stderr,"Read error on file %s\n",fn);
        }
        else
             printf("%s%04x\n","crc = ",aa);
        fclose(fd);
    }
    else {
        fprintf(stderr,"Error - can't open %s\n",fn);
    }
}

main(argc,argv)
int argc; char *argv[];
{
    char *sp,*lip,*fn, *fn2;
    int n,type,attrib;

    sp = malloc(PATHSIZE);
    lip = malloc(PATHSIZE);
    fn = malloc(PATHSIZE);
    fn2 = malloc(PATHSIZE);
    if (argc<=1) {
        fprintf(stderr,"Cyclic Redundancy Check ... by Jon Dart\n");
        fprintf(stderr,"Version %s\n\n",Version);
        fprintf(stderr,"Usage:\n    crc file1 file2 ....\n");
        fprintf(stderr,"    (file specs can be ambiguous)\n");
        exit(1);
    }
    for (n=1;n<argc;n++) {
        type = fixpath(argv[n],sp,lip);
        if (getfirst(sp,0x21,fn,&attrib)) {
            strcpy(fn2,lip);
            strcat(fn2,fn);
            docrc(fn2);
            while (getnext(fn,&attrib)) {
                strcpy(fn2,lip);
                strcat(fn2,fn);
                docrc(fn2);
            }
        }
        else {
            fprintf(stderr,"%s%s\n","Error - can't find: ",argv[n]);
        }
    }
}

