/* this program acts as a filter.  It reads from a file if a file name
is specified, or from the standard input if no file name is given.  The
program converts single carriage returns (^M) or line feeds (^J) to
carriage return/line feed pairs, and writes the resulting text to the
standard output */

/* by Jon Dart, Dept. of Anthropology, UCSD C-001, La Jolla, CA 92093 */
/* version 1.1, 16-Nov-86 */

#include <stdio.h>
#include <fcntl.h>

#define PATHSIZE 	65	/* max size of DOS pathname + 1 */
#define BUFSIZE		8192    /* input buffer size */
#define TRUE		1
#define FALSE		!TRUE
#define CR		'\015'
#define LF		'\012'

extern char *malloc();

struct bufstruct {
	int bc;     /* index into buffer */
	unsigned char *bp;   /* pointer into buffer */
	int nread;  /* number of chars. read on last disk access */
	int fd;	    /* file descriptor */
} bufstuf;

typedef struct bufstruct *bsp;

char nextc(bs)
/* peeks at next char. in buffer. If we're at the very end of the buffer,
   reads ahead to get the next char */
bsp bs;
{
	char buf;	
	int n;

	if (bs->bc < BUFSIZE) 
		return(*(bs->bp));
	else {
		n = read(bs->fd,&buf,1);
		if (n>0) 
			return(buf);
		else
			return('\0');
 	}
}

main(argc,argv)
int argc; char **argv;
{
    int fd;	
    unsigned char *buffer;
    register unsigned char c,lastc;

    buffer = (unsigned char *)  malloc(BUFSIZE);

    if (argc>=2) {
        bufstuf.fd = open(argv[1],O_RDONLY);
        if (bufstuf.fd == -1) {
    	    fprintf(stderr,"\nCan't open: %s\n",argv[1]);
    	    exit(1);
    	}
    }
    else /* read from std input */
        bufstuf.fd = 0;
    bufstuf.bc = BUFSIZE;
    bufstuf.nread = 0;
    lastc = '\0';
    bufstuf.bp = buffer;
    for (;;) {
	if (bufstuf.bc >= bufstuf.nread) {
		if ((bufstuf.nread = read(bufstuf.fd,buffer,BUFSIZE))==0) break;
		bufstuf.bc = 0;
		bufstuf.bp = buffer;
	}
        c = *bufstuf.bp;
        bufstuf.bp++;
        bufstuf.bc++;
        if (c==CR) {
        	putc(c,stdout);
        	if (nextc(&bufstuf)!=LF)
        	    putc(LF,stdout);
        }
        else if (c==LF) {
        	if (lastc!=CR) 
        	    putc(CR,stdout);
        	putc(c,stdout);
        }
        else
           putc(c,stdout);
        lastc = c;
    }
    close(fd);
}
